﻿// Author: Fletcher Baker

using System;
using System.Collections.Generic;
using System.IO;

namespace _2dKnapsack
{
    class Knapsack2d
    {
        private const string _DEFAULT_OUTPUT_PATH = "./output.txt", _DEFAULT_INPUT_PATH = "./input.txt";

        static void Main(string[] args)
        {
            //Set the paths for file I/O to the defaults
            String fin = _DEFAULT_INPUT_PATH, fout = _DEFAULT_OUTPUT_PATH;

            //Set the paths for file I/O to the args if they were passed in
            if (args.Length > 0)
                fin = args[0];
            if (args.Length > 1)
                fout = args[1];

            //Runs the program
            Do(fin, fout);

            //Wait to close until any key is pressed
            Console.WriteLine("<Press any key to close.>");
            Console.ReadKey();
        }

        //Primary driving method of the program
        //Takes the File I/O paths as parameters
        private static void Do(String fin, String fout)
        {
            String[] strArr; //array for holding lines of input or output
            List<Case> cases = null; //List holding each case
            
            strArr = File.ReadAllLines(fin); //Read in all the lines from the input file at the specified path
            cases = ParseInputFile(strArr); //parse the strings into a list of Case objects
            strArr = RunAllCases(cases); //solve the problem, and set strArr to the output of the solution
            PrintOutput(strArr); //prints output to standard out (the console)
            File.WriteAllLines(fout, strArr); //write the output to the file at the specified path
        }

        //prints output to standard out (the console)
        private static void PrintOutput(String[] output)
        {
            foreach (String line in output)
                Console.WriteLine(line);
        }

        //parse the strings into a list of Case objects, so that each case can be handled 
        private static List<Case> ParseInputFile(String[] input)
        {
            List<Case> cases = new List<Case>(); //initiate the list of Case objects to hold the individual cases
            List<Treasure> items; //A temporary list for building the array of treasure types in a case
            int numCases = Int32.Parse(input[0]); //the number of cases in the input; k in the problem statement
            int curr = 1; //the current index in the input array
            int n;  //temporary reference to the number of types of treasure in a case; n in the problem statement
            int x;  //temporary reference to the width of the knapsack; x in the problem statement
            int y;  //temporary reference to the height of the knapsack; y in the problem statement
            int v;  //temporary reference to the value of a treasure; v in the problem statement
            int h;  //temporary reference to the height of a treasure; h in the problem statement
            int w;  //temporary reference to the width of a treasure; w in the problem statement
            char t; //temporary reference to the identifier of a treasure; t in the problem statement
            string[] xy; //string array to temporarily hold the dimensions of the knapsack; x and y in the problem statement
            string[] itemsplit; // string array to temporarily hold all the values from a single treasure

            for (int i = 0; i < numCases; i++) //for each case in the input
            {
                xy = input[curr].Split(' '); //Get the first line of a case and split it into two strings. The first is x, the second is y
                curr++; //increment curr to the next line of input
                n = Int32.Parse(input[curr]); //Get the next line of a case, and assign it to n
                curr++; //increment curr to the next line of input
                items = new List<Treasure>(); //initiate the empty list to temporarily hold all of the treasures for this case
                for (int j = 0; j < n; j++) //for each type of treasure in this case
                {
                    itemsplit = input[curr].Split(' '); //Split the next line into 4 strings; these 4 strings are the 4 values that define a treasure
                    curr++; //increment curr to the next line of input
                    t = itemsplit[0].ToCharArray()[0]; //The first string in the array is t and must be converted to a char
                    v = Int32.Parse(itemsplit[1]); //The second string is v and must be converted to an int
                    w = Int32.Parse(itemsplit[2]); //The third string is w and must be converted to an int
                    h = Int32.Parse(itemsplit[3]); //The forth string is h and must be converted to an int
                    items.Add(new Treasure(t, v, w, h)); //Initiate a new treasure with these values and add it to the list
                }
                x = Int32.Parse(xy[0]); //The first string is x and must be converted to an int
                y = Int32.Parse(xy[1]); //The first string is y and must be converted to an int
                cases.Add(new Case(x, y, n, items.ToArray())); //create a case object with these values and add it to the list
            }
            return cases; //return the list of case objects
        }

        //solve the problem, and set strArr to the output of the solution
        private static String[] RunAllCases(List<Case> cases)
        {
            List<String> output = new List<String>(); //a list to hold all the lines of output

            foreach (Case c in cases) //for each case
                output.Add(RunCase(c)); //add the output from the case to the list

            return output.ToArray(); //return the list as an array
        }

        private static String RunCase(Case c)
        {
            long[,,] benefitTable = new long[c.NumItemTypes + 1, c.Width + 1, c.Height + 1]; //Define the benefit table as you would in the 0/1 Knapsack problem
            Treasure item; //The current item

            for (int i = 1; i <= c.NumItemTypes; i++)   //for each treaure type...
            {                                           //note: started i at 1, because the first row is 0'd out
                for (int w = 1; w <= c.Width; w++)      //for each width allowance...
                {                                       //note: started w at 1, because the first column is 0'd out
                    for (int h = 1; h <= c.Height; h++) //for each height allowance...
                    {                                   //note: started h at 1, because the first row is 0'd out
                        item = c.Treasures[i - 1]; //get the item corresponding to the current row of the benefit table
                        if (item.Height <= h && item.Width <= w) //if the item fits in the current height and width...
                        {
                            benefitTable[i, w, h] = item.Value + benefitTable[i, w, (h - item.Height)] + benefitTable[i, (w - item.Width), h] - benefitTable[i, (w - item.Width), (h - item.Height)]; 
                            //set the current location in the benefit table to the current item's benefit, plus the best benefit that fits in the remaining height
                            if (benefitTable[(i - 1), w, h] > benefitTable[i, w, h]) //If the previous item's benefit is greater than the calculated benefit for the current space...
                                benefitTable[i, w, h] = benefitTable[(i - 1), w, h]; //set the current location in the benefit table equal to the previous item's benefit
                        }
                        else //if the item does not fit in the current height...
                        {
                            benefitTable[i, w, h] = benefitTable[(i - 1), w, h]; //set the current location in the benefit table equal to the previous row's benefit
                        }
                    }
                }
            }
            
            return "Backpack size: " + c.Width + "x" + c.Height + ", Value: " + benefitTable[c.NumItemTypes, c.Width, c.Height]; //return the output for this case
        }

        //A data class for holding all of the info for a single case
        private class Case
        {
            public int Width { get; } //The width of the backpack; x in the problem statement
            public int Height { get; } //The height of the backpack; y in the problem statement
            public int NumItemTypes { get; } //The number of types of treasure; n in the problem statement 
            public Treasure[] Treasures { get; } //An array holding all off types of treasure (listed after n in the input file)

            public Case(int x, int y, int n, Treasure[] items)
            {
                Width = x;
                Height = y;
                NumItemTypes = n;
                Treasures = items;
            }
        }

        //A data class for holding all of the info for a single treasure
        private class Treasure
        {
            public char Type { get; } //The character denoting the treasure type; t in the problem statement
            public int Width { get; } //The width of the treasure; w in the problem statement
            public int Height { get; } //The height of the treasure; h in the problem statement
            public int Value { get; } //The value (or benefit) of the treasure; v in the problem statement

            public Treasure(char t, int v, int w, int h)
            {
                Type = t;
                Value = v;
                Width = w;
                Height = h;
            }
        }
    }
}
